@@ -80,6 +80,8 @@ static struct sigaction saved_sigchld;
 static volatile sig_atomic_t alarm_rised;
 static volatile sig_atomic_t sigchild;
 
+#define SULOGIN_PASSWORD_BUFSIZ	128
+
 #ifndef IUCLC
 # define IUCLC		0
 #endif
@@ -168,16 +170,17 @@ static void tcinit(struct console *con)
 		tio->c_cc[VMIN]     = 1;
 
 		if (ioctl(fd, TIOCGWINSZ, &ws) == 0) {
-			int set = 0;
+			int update = 0;
+
 			if (ws.ws_row == 0) {
 				ws.ws_row = 24;
-				set++;
+				update++;
 			}
 			if (ws.ws_col == 0) {
 				ws.ws_col = 80;
-				set++;
+				update++;
 			}
-			if (set)
+			if (update)
 				ignore_result( ioctl(fd, TIOCSWINSZ, &ws) );
 		}
 
@@ -267,11 +270,11 @@ static void tcfinal(struct console *con)
 		break;
 	case 1:				/* odd parity */
 		tio->c_cflag |= PARODD;
-		/* fall through */
+		/* fallthrough */
 	case 2:				/* even parity */
 		tio->c_cflag |= PARENB;
 		tio->c_iflag |= (INPCK | ISTRIP);
-		/* fall through */
+		/* fallthrough */
 	case (1 | 2):			/* no parity bit */
 		tio->c_cflag &= ~CSIZE;
 		tio->c_cflag |= CS7;
@@ -520,8 +523,9 @@ static void doprompt(const char *crypted, struct console *con, int deny)
 		tty.c_oflag |= (ONLCR | OPOST);
 		tcsetattr(con->fd, TCSADRAIN, &tty);
 	}
-	if (con->file == (FILE*)0) {
-		if  ((con->file = fdopen(con->fd, "r+")) == (FILE*)0)
+	if (!con->file) {
+		con->file = fdopen(con->fd, "r+");
+		if (!con->file)
 			goto err;
 	}
 
@@ -600,13 +604,13 @@ static void setup(struct console *con)
  * Ask for the password. Note that there is no default timeout as we normally
  * skip this during boot.
  */
-static const char *getpasswd(struct console *con)
+static char *getpasswd(struct console *con)
 {
 	struct sigaction sa;
 	struct termios tty;
-	static char pass[128], *ptr;
+	static char pass[SULOGIN_PASSWORD_BUFSIZ], *ptr;
 	struct chardata *cp;
-	const char *ret = pass;
+	char *ret = NULL;
 	unsigned char tc;
 	char c, ascval;
 	int eightbit;
@@ -643,7 +647,7 @@ static const char *getpasswd(struct console *con)
 				xusleep(250000);
 				continue;
 			}
-			ret = (char*)0;
+			ret = NULL;
 			switch (errno) {
 			case 0:
 			case EIO:
@@ -695,7 +699,7 @@ static const char *getpasswd(struct console *con)
 		default:
 			if ((size_t)(ptr - &pass[0]) >= (sizeof(pass) -1 )) {
 				 fprintf(stderr, "sulogin: input overrun at %s\n\r", con->tty);
-				 ret = (char*)0;
+				 ret = NULL;
 				 goto quit;
 			}
 			*ptr++ = ascval;
@@ -709,6 +713,10 @@ quit:
 	tcfinal(con);
 	printf("\r\n");
 out:
+#ifdef HAVE_EXPLICIT_BZERO
+	if (ret == NULL)
+		explicit_bzero(pass, sizeof(pass));
+#endif
 	return ret;
 }
 
@@ -773,9 +781,10 @@ static void sushell(struct passwd *pwd)
 
 #ifdef HAVE_LIBSELINUX
 	if (is_selinux_enabled() > 0) {
-		security_context_t scon=NULL;
-		char *seuser=NULL;
-		char *level=NULL;
+		char *scon = NULL;
+		char *seuser = NULL;
+		char *level = NULL;
+
 		if (getseuserbyname("root", &seuser, &level) == 0) {
 			if (get_default_context_with_level(seuser, level, 0, &scon) == 0) {
 				if (setexeccon(scon) != 0)
@@ -787,16 +796,17 @@ static void sushell(struct passwd *pwd)
 		free(level);
 	}
 #endif
-	execl(su_shell, shell, NULL);
+	execl(su_shell, shell, (char *)NULL);
 	warn(_("failed to execute %s"), su_shell);
 
 	xsetenv("SHELL", "/bin/sh", 1);
-	execl("/bin/sh", profile ? "-sh" : "sh", NULL);
+	execl("/bin/sh", profile ? "-sh" : "sh", (char *)NULL);
 	warn(_("failed to execute %s"), "/bin/sh");
 }
 
-static void usage(FILE *out)
+static void usage(void)
 {
+	FILE *out = stdout;
 	fputs(USAGE_HEADER, out);
 	fprintf(out, _(
 		" %s [options] [tty device]\n"), program_invocation_short_name);
@@ -811,9 +821,10 @@ static void usage(FILE *out)
 		out);
 
 	fputs(USAGE_SEPARATOR, out);
-	fputs(USAGE_HELP, out);
-	fputs(USAGE_VERSION, out);
-	fprintf(out, USAGE_MAN_TAIL("sulogin(8)"));
+	printf(USAGE_HELP_OPTIONS(26));
+	printf(USAGE_MAN_TAIL("sulogin(8)"));
+
+	exit(EXIT_SUCCESS);
 }
 
 int main(int argc, char **argv)
@@ -842,7 +853,7 @@ int main(int argc, char **argv)
 	INIT_LIST_HEAD(&consoles);
 
 	/*
-	 * If we are init we need to set up a own session.
+	 * If we are init we need to set up an own session.
 	 */
 	if ((pid = getpid()) == 1) {
 		setsid();
@@ -852,7 +863,7 @@ int main(int argc, char **argv)
 	setlocale(LC_ALL, "");
 	bindtextdomain(PACKAGE, LOCALEDIR);
 	textdomain(PACKAGE);
-	atexit(close_stdout); /* XXX */
+	close_stdout_atexit();
 
 	/*
 	 * See if we have a timeout flag.
@@ -869,14 +880,11 @@ int main(int argc, char **argv)
 			opt_e = 1;
 			break;
 		case 'V':
-			printf(UTIL_LINUX_VERSION);
-			return EXIT_SUCCESS;
+			print_version(EXIT_SUCCESS);
 		case 'h':
-			usage(stdout);
-			return EXIT_SUCCESS;
+			usage();
 		default:
-			usage(stderr);
-			/* Do not exit! */
+			/* Do not exit! getopt prints a warning. */
 			break;
 		}
 	}
@@ -971,21 +979,24 @@ int main(int argc, char **argv)
 		switch ((con->pid = fork())) {
 		case 0:
 			mask_signal(SIGCHLD, SIG_DFL, NULL);
-			/* fall through */
 		nofork:
 			setup(con);
 			while (1) {
 				const char *passwd = pwd->pw_passwd;
-				const char *answer;
-				int failed = 0, doshell = 0;
+				char *answer;
+				int doshell = 0;
 				int deny = !opt_e && locked_account_password(pwd->pw_passwd);
 
 				doprompt(passwd, con, deny);
 
 				if ((answer = getpasswd(con)) == NULL)
 					break;
-				if (deny)
+				if (deny) {
+#ifdef HAVE_EXPLICIT_BZERO
+					explicit_bzero(answer, SULOGIN_PASSWORD_BUFSIZ);
+#endif
 					exit(EXIT_FAILURE);
+				}
 
 				/* no password or locked account */
 				if (!passwd[0] || locked_account_password(passwd))
@@ -998,17 +1009,17 @@ int main(int argc, char **argv)
 					else if (strcmp(cryptbuf, pwd->pw_passwd) == 0)
 						doshell++;
 				}
-
+#ifdef HAVE_EXPLICIT_BZERO
+				explicit_bzero(answer, SULOGIN_PASSWORD_BUFSIZ);
+#endif
 				if (doshell) {
+					/* sushell() unmask signals */
 					sushell(pwd);
-					failed++;
-				}
 
-				mask_signal(SIGQUIT, SIG_IGN, &saved_sigquit);
-				mask_signal(SIGTSTP, SIG_IGN, &saved_sigtstp);
-				mask_signal(SIGINT,  SIG_IGN, &saved_sigint);
+					mask_signal(SIGQUIT, SIG_IGN, &saved_sigquit);
+					mask_signal(SIGTSTP, SIG_IGN, &saved_sigtstp);
+					mask_signal(SIGINT,  SIG_IGN, &saved_sigint);
 
-				if (failed) {
 					fprintf(stderr, _("cannot execute su shell\n\n"));
 					break;
 				}
@@ -1024,7 +1035,7 @@ int main(int argc, char **argv)
 			exit(0);
 		case -1:
 			warn(_("fork failed"));
-			/* fall through */
+			/* fallthrough */
 		default:
 			break;
 		}
