@@ -43,6 +43,7 @@
 #include "optutils.h"
 
 static int quiet;
+static struct ul_env_list *envs_removed;
 
 static int table_parser_errcb(struct libmnt_table *tb __attribute__((__unused__)),
 			const char *filename, int line)
@@ -117,19 +118,21 @@ static void suid_drop(struct libmnt_context *cxt)
 	const uid_t ruid = getuid();
 	const uid_t euid = geteuid();
 
-	if (ruid != 0 && euid == 0) {
-		if (setgid(getgid()) < 0)
-			err(MNT_EX_FAIL, _("setgid() failed"));
-
-		if (setuid(getuid()) < 0)
-			err(MNT_EX_FAIL, _("setuid() failed"));
-	}
+	if (ruid != 0 && euid == 0 && drop_permissions() != 0)
+		err(MNT_EX_FAIL, _("drop permissions failed"));
 
 	/* be paranoid and check it, setuid(0) has to fail */
 	if (ruid != 0 && setuid(0) == 0)
 		errx(MNT_EX_FAIL, _("drop permissions failed."));
 
 	mnt_context_force_unrestricted(cxt);
+
+	/* restore "bad" environment variables */
+	if (envs_removed) {
+		env_list_setenv(envs_removed);
+		env_list_free(envs_removed);
+		envs_removed = NULL;
+	}
 }
 
 static void success_message(struct libmnt_context *cxt)
@@ -291,13 +294,20 @@ static int umount_one_if_mounted(struct libmnt_context *cxt, const char *spec)
 static int umount_do_recurse(struct libmnt_context *cxt,
 		struct libmnt_table *tb, struct libmnt_fs *fs)
 {
-	struct libmnt_fs *child;
+	struct libmnt_fs *child, *over = NULL;
 	struct libmnt_iter *itr = mnt_new_iter(MNT_ITER_BACKWARD);
 	int rc;
 
 	if (!itr)
 		err(MNT_EX_SYSERR, _("libmount iterator allocation failed"));
 
+	/* first try overmount */
+	if (mnt_table_over_fs(tb, fs, &over) == 0 && over) {
+		rc = umount_do_recurse(cxt, tb, over);
+		if (rc != MNT_EX_SUCCESS)
+			goto done;
+	}
+
 	/* umount all children */
 	for (;;) {
 		rc = mnt_table_next_child_fs(tb, itr, fs, &child);
@@ -309,6 +319,9 @@ static int umount_do_recurse(struct libmnt_context *cxt,
 		} else if (rc == 1)
 			break;		/* no more children */
 
+		if (over && child == over)
+			continue;
+
 		rc = umount_do_recurse(cxt, tb, child);
 		if (rc != MNT_EX_SUCCESS)
 			goto done;
@@ -416,7 +429,7 @@ done:
 
 /*
  * Check path -- non-root user should not be able to resolve path which is
- * unreadable for him.
+ * unreadable for them.
  */
 static char *sanitize_path(const char *path)
 {
@@ -486,7 +499,7 @@ int main(int argc, char **argv)
 	};
 	int excl_st[ARRAY_SIZE(excl)] = UL_EXCL_STATUS_INIT;
 
-	sanitize_env();
+	__sanitize_env(&envs_removed);
 	setlocale(LC_ALL, "");
 	bindtextdomain(PACKAGE, LOCALEDIR);
 	textdomain(PACKAGE);
@@ -630,6 +643,8 @@ int main(int argc, char **argv)
 	}
 
 	mnt_free_context(cxt);
+	env_list_free(envs_removed);
+
 	return (rc < 256) ? rc : 255;
 }
 
